#ifndef _zrp_h_
#define _zrp_h_

#include <config.h>
#include <assert.h>
#include <agent.h>
#include <packet.h>
#include <ip.h>
#include <delay.h>
#include <scheduler.h>
#include <queue.h>
#include <trace.h>
#include <arp.h>
#include <ll.h>
#include <mac.h>
#include <priqueue.h>
#include <delay.h>
#include "zrp_table.h"

#if defined(WIN32) && !defined(snprintf)
#define snprintf _snprintf
#endif /* WIN32 && !snprintf */

#define ROUTER_PORT      0xff
#define FOUND 1
#define LOST 0
#define TRUE 1
#define FALSE 0
#define NULL 0

#define MAX_OUTBOUND_PENDING 20 // max queue for pkts waiting for ierp requests
#define DEFAULT_ZONE_RADIUS 3 

#define MAX_DETECTED_QUERIES 20 // max queries that can be detected 
#define IERP_REQUEST_TIMEOUT 30 // secs before timeout on an ierp request
#define ZRP_STARTUP_JITTER 2.0	// secs to jitter start of periodic activity
#define MAX_SEQUENCE_ID 1000000000
#define QUERY_TABLE_EXPIRATION 1000
#define MAX_QUERY_DETECT_TABLE_SIZE 50

#define NEIGHBOR_ACK_TIMEOUT 10 // sec
#define TIMER_JITTER 3

#define BEACON_PERIOD 1  // period of beacon transmission in sec
#define NDP_SCAN_PERIOD 5
        // scan NDP table every number of beacon timeout cycles 
#define MAX_NODE_ID_CHAR_SIZE 10 // longest name string for my node address
                             // (char *)myid_ is string equiv of (int)myaddr_ 
#define MAX_IERP_CACHE    50  // max ierp requests cache size
#define MAX_QID           50  // max size query detect table
#define FOUND 1
#define LOST 0
#define TRUE 1
#define FALSE 0

typedef double Time;   
typedef int32_t Query_ID;

class ZRPAgent;

enum ZRPTYPE // Types of ZRP packets
{
  NDP_BEACON, NDP_BEACON_ACK, IARP_UPDATE, IARP_DATA, IERP_REPLY, IERP_REQUEST 
};

// link state data used in updates
struct linkie {
  
  int src_;   // 32 bits
  int dest_;  // 32 bits
  int isup_;
};

// ZRP header structure
struct hdr_zrp {
  int zrptype_;
  //  int safe_;
  int routeindex_;
    
  // IARP hdr
  int seq_;
  int numentries_; // number of link entries in this update
  // or length of direct routing list
  // info is in data portion of packet, use accessdata()

  nsaddr_t originator_; // for ndp, iarp or ierp
  nsaddr_t ierpsource_;
  nsaddr_t ierpdest_;
  nsaddr_t iarpsource_;
  nsaddr_t iarpdest_;
  nsaddr_t lastbordercaster_;
  Time pktsent_;
    
  // this is where the original data for upper layer pkts
  // is saved while ZRP routes pkt, at dest this is placed
  // back into hdrip->dport(), ie this is part of encapsulated data
  int enc_dport_;
  int enc_daddr_;
  packet_t enc_ptype_;
  
  int qid_;  // query id counter

  nsaddr_t *route_; // pointer to route list data
  linkie *links_;  // pointer to link state list 
  
  int forwarded_;  // TRUE if forwarded before
  
  // Packet header access functions
  static int offset_;
  inline static int& offset() { return offset_; }
  inline static hdr_zrp* access(const Packet* p) {
    return (hdr_zrp*) p->access(offset_);
  }
  
};

class ZRPBeaconTransmitTimer : public Handler {
 public:
  
  ZRPBeaconTransmitTimer(ZRPAgent* a)  { agent = a; } 
  // pgi- set zrp_agent to 'a', ie our ZRP_Agent
  void handle(Event*);  // function handling the event
  void start(double thistime);
  
  //  private:
  ZRPAgent *agent;
  Event intr;
  
};

class ZRPNeighborScanTimer : public Handler {
 public:
  
  ZRPNeighborScanTimer(ZRPAgent* a)  { agent = a; } 
  // pgi- set zrp_agent to 'a', ie our ZRP_Agent
  void handle(Event*);  // function handling the event
  void start(double thistime);
  
  //  private:
  ZRPAgent *agent;
  Event intr;
  
};

class ZRPPeriodicUpdateTimer : public Handler {
 public:
  
  ZRPPeriodicUpdateTimer(ZRPAgent* a)  { agent = a; } 
  // pgi- set zrp_agent to 'a', ie our ZRP_Agent
  void handle(Event*);  // function handling the event
  void start(double thistime);
    
  //  private:
  ZRPAgent *agent;
  Event intr;
  
};

class ZRPAckTimer : public Handler {
 public:
  
  ZRPAckTimer(ZRPAgent* a)  { agent = a; } 
  // pgi- set zrp_agent to 'a', ie our ZRP_Agent
  void handle(Event*);  // function handling the event
  void start();
  
  //  private:
  ZRPAgent *agent;
  Event intr;
  
};


class ZRPAgent : public Agent {

  friend class ZRPBeaconTimer;
 public:
  ZRPAgent(); // not generally used
  ZRPAgent(nsaddr_t id);
  
  void recv(Packet * p, Handler *);
  int  initialized() { return 1 && target_; } 

  //== Tcl Related ======================================
  char*  myid_;   // (char *)myid_ is string equiv of (int)myaddr_ 
  PriQueue* ll_queue;
  Trace* tracetarget;
  MobileNode* node_;
  NsObject* port_dmux_;
  
  //==Timer Related ==========================================

  ZRPNeighborScanTimer NeighborScanTimer_;
  ZRPBeaconTransmitTimer BeaconTransmitTimer_;
  ZRPPeriodicUpdateTimer PeriodicUpdateTimer_;
  ZRPAckTimer AckTimer_;
    
  //==Methods ================================================
  
  void startup();
  int  command (int argc, const char*const* argv);
  void scan_ierp_cache();
  void ierp_notify_iarp(Packet *p);
  void recvZRP(Packet* p);
  void forward_update(Packet* p);
  void route_pkt(Packet* p, nsaddr_t dest);
  void print_tables();
  
  int  neighbor_add(nsaddr_t newneighbor);
  void send_periodic_update();
  void sendp(nsaddr_t ns);
  void originate_request(nsaddr_t final, int qid);
  void scan_query_detections();
  void scan_linkstate_table();
  int query_detect1(int qid, nsaddr_t originid, nsaddr_t lastbordercaster,
		    int radius);
  int  node_is_covered(int queryid, nsaddr_t node);
  void inc_seq() { seq_++; if (seq_ > MAX_SEQUENCE_ID) seq_ = 1;}
  void dec_seq() { seq_--; if (seq_ < 1)  seq_ = MAX_SEQUENCE_ID;} 
  void pkt_add_all_links(Packet *p);
  void pkt_send(Packet *p, nsaddr_t addressee);
  void send_next_hop(Packet *p);
  void send_prev_hop(Packet *p);
  void pkt_broadcast(Packet *p);
  Packet* pkt_create(ZRPTYPE zrp_type, nsaddr_t addressee, int ttl);
  void pkt_add_node(Packet* p, nsaddr_t node, int total);
  void pkt_add_link(Packet* p, nsaddr_t lnksrc, nsaddr_t lnkdest, int isup);
  void pkt_read_update(Packet* p, link_struct *ls, int index) ;
  nsaddr_t find_next_hop(Packet* p);
  nsaddr_t find_prev_hop(Packet* p);
  int add_local_route(Packet* p, nsaddr_t node);
  int add_outer_route(Packet* p, nsaddr_t node);
  void do_update();
  void print_routes();
  void pkt_create_route_data_space(Packet *p, int size);
  void pkt_free_route_data_space(Packet *p);
  void pkt_create_link_data_space(Packet *p, int size);
  void pkt_free_link_data_space(Packet *p);
  void pkt_copy(Packet *pfrom, Packet *pto);
  void zdrop(Packet *p, const char *s);
  int  node_is_covered(nsaddr_t node, int queryid, nsaddr_t originid);
  void outer_route_expiration();
  void IARP_update_outer_route(int index);
    
  //==Data ===================================================

  int myaddr_;
  int radius_;   // Can be set from Tcl script, see ZRPAgent::command
  int transmit_jitter_;
  int startup_jitter_;
  int process_jitter_;
  int beacon_period_; // Can be set from Tcl script, see ZRPAgent::command
  int beacon_period_jitter_;
  int neighbor_timeout_;
  int neighbor_ack_timeout_;
  int iarp_update_period_;
  int neighbortable_scan_period_;
  
  int num_neighbors_;
  
  int tx_; // total pkts transmitted by agent
  int rx_; // total pkts received by agent
  int seq_; // current sequence id for outgoing IARP updates
  int qid_; // global query id counter, updated for every query sent by node
  
  int suspend_flag_; // Can be set from Tcl script, see ZRPAgent::command
  // "suspends" node, there are Tcl commands that work better "start","stop"

  //==Tables ===================================================
  neighbor_struct *neighbors_; // linked list of my neighbors
  LinkStateTable lstable;     // class, linked list
  RoutingTable rtable;        // class, array of pointers
  IERPCacheEntry *ierpcache[MAX_IERP_CACHE]; // array of pointers
  QueryDetectEntry *queries_[MAX_QUERY_DETECT_TABLE_SIZE];
};

#endif
