
#include <config.h>
#include <agent.h>
#include <packet.h>
#include <ip.h>
#include <delay.h>
#include <scheduler.h>
#include <queue.h>
#include <trace.h>
#include <arp.h>
#include <ll.h>
#include <mac.h>
#include <priqueue.h>
#include <delay.h>
#include <random.h>
#include <object.h>
#include <route.h>
#include "zrp.h"
#include "zrp_table.h"

#define UP    1
#define DOWN  0

//===========================================================================
// Linkstate Table:
// Perform an update to table given a link state, Return seq_match=TRUE
// if there was a match. Return update=TRUE if any change in table.
//===========================================================================
BOOLEAN
LinkStateTable::update(int src, int dest, int lstate, int seq, int *seq_match) 
{
  Time now = Scheduler::instance().clock(); // get the time
  int flag; // TRUE indicates there has been a change in table
  link_struct* i;
  int found;

  *seq_match = FALSE;
  found = FALSE;
  flag = FALSE;
  
  if  (lshead_ == NULL) { // check if empty
    if (lstate == UP) { // add only if up
      // add to table, 
      lshead_ = new link_struct;
      lshead_->src_ = src;
      lshead_->dest_ = dest;
      lshead_->isup_ = lstate;
      lshead_->seq_ = seq;
      lshead_->expiry_ = now + LINKSTATE_EXPIRATION;
      lshead_->next_ = NULL;
      return TRUE;  // change in table
    }
    return FALSE;
  }
  
  for (i=lshead_; i != NULL; i=i->next_) {
    if ( ( (i->src_ == src)  && (i->dest_ == dest) ) || ( (i->src_ == dest) && (i->dest_ == src)  ) ) {  // src/dest match with update

      if ( (seq != 0) && (i->seq_ >= seq) ) {  // got this update before
	*seq_match = TRUE;
	return FALSE;
      }
      
      found = TRUE; // found a match (a,b) or (b,a)

      i->expiry_ = now + LINKSTATE_EXPIRATION; // update expiry
      
      if (seq != 0 )  // only update if new seq is non-zero (
	i->seq_ = seq; // update to new seq no.

      if ( i->isup_ != lstate) { //inequality=we need to change state to latest
	i->isup_ = lstate;
	flag = TRUE;
      }
    }
  }

  if (found) // return flag if there was change
    return flag;

  // we have found nothing up until this point, so add link
  
  if (lstate == UP) {  // don't want to add a downed link
    // add to table, 
    i = new link_struct;
    i->src_ = src;
    i->dest_ = dest;
    i->isup_ = lstate;
    i->seq_ = seq;
    i->next_ = lshead_;
    i->expiry_ = now + LINKSTATE_EXPIRATION;
    lshead_ = i;
    return TRUE;  // change in table    
  }
  return FALSE;
}

//=======================================================================
// Linkstate Table:
// Insert Link state entry into Link state table
//=======================================================================
void  
LinkStateTable::add(link_struct* link)
{  
  // insert link at head
  if (lshead_ != NULL) {  // if there is at least one link in list
    link->next_ = lshead_;       ;
    lshead_ = link;
  } else { // if empty
    lshead_ = link;
    lshead_->next_ = NULL;
  }

}

//========================================================================
// Route Table:
// Erases route and peripheral tables.
//========================================================================
void RoutingTable::erase() {
  // would be better if we saved fallow data structures 
  // on a free list - for version 2.0 ;)
  int i;
  rtable_node *tmp;

  //erase route table
  for (i=0; (i< MAX_NODES_PER_ZONE) && (route[i] != NULL) ; i++) {
    delete(route[i]);
    route[i] = NULL;
  } //end for
  
  // erase periph node list 
  while (periphnodes_ != NULL ) {
    tmp = periphnodes_;
    periphnodes_ = periphnodes_->next_;
    delete(tmp);
  }
  periphnodes_ = NULL;
}

//=========================================================================
// Linkstate Table:
// Clean table of expired links.
//==========================================================================
void
LinkStateTable::purge_links() 
{
  Time now = Scheduler::instance().clock(); // get the time
  link_struct *ls;
  
  // deletes down or expired links

  if (lshead_ == NULL) 
    return;

  // hack, insert a dummy node at head
  link_struct dummy;  
  dummy.next_ = lshead_;
  lshead_ = &dummy;
  
  for (ls=lshead_ ; ls->next_ != NULL; ) {

    if ( (ls->next_->expiry_ < now) || (ls->next_->isup_ == FALSE ) ) {
      link_struct *temp = ls->next_;
      ls->next_ = ls->next_->next_;
      delete temp;
      // other actions to take when you delete a node here
    } else {
      ls=ls->next_ ;
    }
        
  }
  lshead_ = dummy.next_;
  
}

//===========================================================================
// Linkstate Table:
// Print table to STDOUT.
//===========================================================================
void
LinkStateTable::print_links() 
{
  link_struct *ls;
  
  if (lshead_ == NULL) {
    printf("LinkTable: empty ");
  } else {
    printf("LinkTable: ");  
    
    for (ls = lshead_;  
	 (ls != NULL);
       ls=ls->next_) {   // If link list has one or more link
          
      if (ls->isup_) {
	printf("%d=%d ",ls->src_,ls->dest_); 
      } else {
	printf("%d#%d ",ls->src_,ls->dest_); 
      }
    } // for link_stuct 
  }
}

//===========================================================================
// Route Table:
// Print entries.
//===========================================================================
void 
RoutingTable::print() {
  rtable_node *j;
  int i;
  Time now = Scheduler::instance().clock(); // get the time
  printf("_%2d_ [%6.6f] | Route Table ",my_address_,now);
   
  for (i=0; route[i] != NULL && i< MAX_NODES_PER_ZONE ; i++) {
    
    printf("| Entry %d : ",i);
    
    if (route[i]==NULL) {
      printf("empty");
      return;
    } 
    
    for (j=route[i]; j!=NULL; j=j->next_  ) {
      printf("%d(%d)->",(int)j->id_,j->numhops_);
    } 
  }
  if (i==0) {
    printf("empty "); 
  }

}

//===========================================================================
// Route Table:
// Compute minimum hop routes to all nodes that are "radius" hops or less
// with respect to node myadd, given link state table. 
//===========================================================================
void
RoutingTable::compute_routes(nsaddr_t myadd, int radius) {
  link_struct *lptr; // a temp ptr for iterative loops
  link_struct *lst;  // a copy of link table
  link_struct* lnk;  // a ptr to newly created links
  link_struct* ls;   // a temp ptr for iterative loops
  int i;             // counter for iterative loop over route table
  int flag;
  Time now = Scheduler::instance().clock(); // get the time      
  rtable_node *tmp;
  
  flag = 0;
  lst = NULL;
  erase(); // route table
    
  if (linkstatetable->lshead_ == NULL) {    // If link list is empty
    return; // nothing broken, just ran out of link states
  }
    
  // copy valid links in linkstatetable list to lst
  for (lptr=linkstatetable->lshead_; lptr!=NULL; lptr=lptr->next_) { 
    flag = 1;
    
    if (lptr->isup_) {  // if link is valid, add to lst
      lnk = new link_struct(lptr->src_,lptr->dest_,TRUE,lst);
      lnk->src_ = lptr->src_;
      lnk->dest_ = lptr->dest_;
      lnk->isup_ = TRUE;
      // insert link at head of lst
      if (lst != NULL) { 
	lnk->next_ = lst;     
	lst = lnk;
      } else { // else if lst is empty
	lst = lnk;
	lst->next_ = NULL;
      } // end else
    }  // end if
    
  } // end while
  if (!flag)  { }

  route[0] = new rtable_node();
  route[0]->id_= myadd;
  route[0]->next_ = NULL;
  route[0]->numhops_ = 0;
  
  route[1] = NULL;

  num_entries_=1;
  
  // rtable[i] is "current" src we are looking at
  // search for all its neighbors dest(i), which gives us
  // a number of (current src,dest(i) pairs
  // for each pair found, check if dest is in route table
  // if not, add to route table. delete link from lst (set isup_=FALSE)
  // after you check the dest. 
  // add-drop will add/drop entries in route table
  for ( i=0; (i < MAX_NODES_PER_ZONE) && (route[i] != NULL); i++) {
  
    for (ls=lst;  // lst is a copy of linkstatetable, from above  
	 (ls != NULL);
	 ls=ls->next_)
      {
	if (ls->isup_ == TRUE) { // make sure it hasn't been "dropped"
	  add_drop(ls,i);
	}
      } // end for ls
  } // end for i

  // calculate peripheral nodes
  for (i=0; (i < MAX_NODES_PER_ZONE) && (route[i] != NULL); i++) {
    if (route[i]->numhops_ == radius ){
      rtable_node *temp = new rtable_node;
      temp->id_ = route[i]->id_;
      temp->next_ = periphnodes_;
      periphnodes_ = temp;
    }
  }
  empty_links(lst); 
  return;
}

//===========================================================================
// Route Table:
// Looks up node with address "id" in route table, returns index into table
// if it exists, otherwise sets to NotInTable.
//===========================================================================
int
RoutingTable::get_entry(nsaddr_t id) 
{
  int i;
  for (i=0;((route[i] != NULL) && (i<MAX_NODES_PER_ZONE+1)); i++) 
    if (route[i]->id_ == id) 
      return i;
  return (MAX_NODES_PER_ZONE+1);
}

//===========================================================================
// Route Table:
// Looks up node with address "id" in route table, returns TRUE if in table,
// otherwise false.
//===========================================================================
int
RoutingTable::route_exists(nsaddr_t id) 
{
 int i;
  for (i=0;((route[i] != NULL) && (i<MAX_NODES_PER_ZONE+1)); i++) 
    if (route[i]->id_ == id) 
      return TRUE;
  return FALSE;
}

//===========================================================================
// Linkstate Table:
// Delete out lstates in link state table.
//===========================================================================
void
RoutingTable::empty_links(link_struct *lst) 
{
  link_struct *lnk;
  while (lst != NULL) {
    lnk = lst;
    lst = lst->next_;
    delete lnk;
  }
}

//===========================================================================
// Route Table:
// Looks up node with address "nodeid" in route table,
// returns index if in table, otherwise -1.
//===========================================================================
int
RoutingTable::node_exists(nsaddr_t nodeid)
{
  int i;
  for (i=0;((route[i] != NULL) && (i<MAX_NODES_PER_ZONE+1)); i++) 
    if (route[i]->id_ == nodeid) 
      return i; // found
  return (-1); // not found
}

//============================================================================
// Route Table:
// Used by compute_routes() to calculate minimum hop tree.
// Add one more entry to route tree if this node "entree" is found to be
// one hop from an entry already in route tree, ie if node is equal to
// src or dest of this link, "lnk". If a match, this link is marked
// as covered for next pass. If no match, return, caller will continue
// iterating through link state table.
//============================================================================
void
RoutingTable::add_drop(link_struct *lnk, int entree)
{ 

  // if link is not up, nothing left to do
  if (lnk->isup_ == FALSE) { 
    return;
  }

  // if src is in route table
  if ( lnk->src_ == route[entree]->id_ ) {  
    // found a link with same src
    
    // see if dest exists in route table
    if (get_entry(lnk->dest_) != (MAX_NODES_PER_ZONE+1)) {
      
      // yes route already exists, don't add
      lnk->isup_ = FALSE; // in effect dropping link   
      //return TRUE; // but do delete current link_struct
      return;
      
    } else if ( route[ get_entry(lnk->src_) ]->numhops_ < agent_->radius_ ) {
      // add node to RouteTable if within radius
      
      if (num_entries_ == MAX_NODES_PER_ZONE) {
	// error, overflow
	printf("Need to increase MAX_NODES_PER_ZONE in zrp.h to greater"
	       "than %d\n",MAX_NODES_PER_ZONE);
      }
      // add a new route to route table, id=dest,
      // next is route entry with id=src
      route[num_entries_] = new rtable_node;
      route[num_entries_]->id_ = lnk->dest_;
      route[num_entries_]->next_ = route[ get_entry(lnk->src_) ];
      if (route[num_entries_]->next_) {
	route[num_entries_]->numhops_ = route[num_entries_]->next_->numhops_+1;
      } else {
	route[num_entries_]->numhops_ = 0;
      }
      num_entries_++;
      lnk->isup_ = FALSE; // in effect dropping link   
    }
    //return TRUE; // tell the caller to delete the current link_struct
    return;
    
  }  else if ( lnk->dest_ == route[entree]->id_ ) {  
    // found a link with same dest
    
    // see if src exists in route table
    if (get_entry(lnk->src_) != (MAX_NODES_PER_ZONE+1)) {
      
      // yes route already exists, don't add
      lnk->isup_ = FALSE; // in effect dropping link   
      //return TRUE; // but do delete current link_struct
      return;

      // if dest is in route table
    } else if ( route[ get_entry(lnk->dest_) ]->numhops_ < agent_->radius_ ) {

      if (num_entries_ == MAX_NODES_PER_ZONE) {
	// error, overflow
	printf("Need to increase MAX_NODES_PER_ZONE in zrp.h to greater"
	       "than %d\n",MAX_NODES_PER_ZONE);
      }
      // add a new route to route table, id=dest,
      //next is route entry with id=src
      route[num_entries_] = new rtable_node;
      route[num_entries_]->id_ = lnk->src_;
      route[num_entries_]->next_ = route[ get_entry(lnk->dest_) ];
      if (route[num_entries_]->next_) {
	route[num_entries_]->numhops_ = route[num_entries_]->next_->numhops_+1;
      } else {
	route[num_entries_]->numhops_ = 0;
      }
      num_entries_++;
      lnk->isup_ = FALSE; // in effect dropping link   
    }
    return;
  }
  return;
}
